/* ZJ Wood for CSC 476 Spring 2007
  Example program which includes both multi-texturing and a pixel shader for using a diffuse light map 
  along with a decal texture - Multi-texturing code is based of ATI's example light mapping code and the
  vertex and pixel shader integration are based off examples from Nvidia's Cg example code (#7 two texture example)
  Use "l" to toggle the two modes (shader currently includes a bumped up red value for differentiation purposes only!
  Note that you will need the texture examples and .cg files for this to work - ALSO this code was compiled and
  tested on a Macintosh - you will need to change header directories for a PC - May contain random unused code*/


#include <Cg/cg.h>						//Cg Header
#include <Cg/cgGL.h>					//Cg OpenGL Specific Header
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <GLUT/glut.h>
#include <OPENGL/gl.h>
#include <OPENGL/glext.h>
#include <assert.h>
#include <map>

#include "tex.h"

/* Vector3D struct mostly used only for camera nav */
typedef struct Vector3 {
  float x;
  float y;
  float z;
  
  float nx;
  float ny;
  float nz;
  Vector3(float in_x, float in_y, float in_z) : x(in_x), y(in_y), z(in_z) {}
  Vector3() : x(0), y(0), z(0) {}
  Vector3(const Vector3& in_v) : x(in_v.x), y(in_v.y), z(in_v.z) {}
  Vector3(float in) : x(in), y(in), z(in) {}
  inline Vector3& operator =(const Vector3& v2) { x = v2.x; y=v2.y; z=v2.z; return *this;}
  inline void set(float in_x, float in_y, float in_z) { x = in_x; y = in_y; z = in_z;}
} Vector3;

//globals used mostly for camera
Vector3 eye, la;
int GW;
int GH;
int mouseX;
int mouseY;
int startx, starty, endx, endy;
float phi, theta;

void display();

/* Cg variables */
static CGcontext   myCgContext;
static CGprofile   myCgVertexProfile,
                   myCgFragmentProfile;
static CGprogram   myCgVertexProgram,
                   myCgFragmentProgram;

static const char *myProgramName = "p4mac",
                  *myVertexProgramFileName = "lmap_v.cg",
     *myVertexProgramName = "lmap_v",
                  *myFragmentProgramFileName = "lmap_f.cg",
     *myFragmentProgramName = "l_map";

// The Parameters Needed For Our Shader
CGparameter	position, texCoord, oPosition, oTexCoord, modelViewProj, lightMapTexCoord, color, lightMap, decalMap;

/* Nice function for checkign and reporting Cg errors */
static void checkForCgError(const char *situation)
{
  CGerror error;
  const char *string = cgGetLastErrorString(&error);

  if (error != CG_NO_ERROR) {
    printf("%s: %s: %s\n",
      myProgramName, situation, string);
    if (error == CG_COMPILER_ERROR) {
      printf("%s\n", cgGetLastListing(myCgContext));
    }
    exit(1);
  }
}	
										
/* Initialization for some OpenGL stuff and for Cg stuff */
bool Initialize ()					
{

	// Start Of User Initialization
	glClearColor (0.0f, 0.0f, 0.0f, 0.5f);								// Black Background
	glClearDepth (1.0f);												// Depth Buffer Setup
	glDepthFunc (GL_LEQUAL);											// The Type Of Depth Testing
	glEnable (GL_DEPTH_TEST);											// Enable Depth Testing
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);					// Set Perspective Calculations To Most Accurate
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);							// Draw polygons filled in

    /*Cg initialization code */
	  myCgContext = cgCreateContext();
	  checkForCgError("creating context");

	  myCgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
	  cgGLSetOptimalOptions(myCgVertexProfile);
	  checkForCgError("selecting vertex profile");

	  myCgVertexProgram =
		cgCreateProgramFromFile(
		  myCgContext,              /* Cg runtime context */
		  CG_SOURCE,                /* Program in human-readable form */
		  myVertexProgramFileName,  /* Name of file containing program */
		  myCgVertexProfile,        /* Profile: OpenGL ARB vertex program */
		  myVertexProgramName,      /* Entry function name */
		  NULL);                    /* No extra compiler options */
	  checkForCgError("creating vertex program from file");
	  cgGLLoadProgram(myCgVertexProgram);
	  checkForCgError("loading vertex program");

	  myCgFragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	  cgGLSetOptimalOptions(myCgFragmentProfile);
	  checkForCgError("selecting fragment profile");

	  myCgFragmentProgram =
		cgCreateProgramFromFile(
		  myCgContext,                /* Cg runtime context */
		  CG_SOURCE,                  /* Program in human-readable form */
		  myFragmentProgramFileName,  /* Name of file containing program */
		  myCgFragmentProfile,        /* Profile: OpenGL ARB vertex program */
		  myFragmentProgramName,      /* Entry function name */
		  NULL);                      /* No extra compiler options */
	  checkForCgError("creating fragment program from file");
	  cgGLLoadProgram(myCgFragmentProgram);
	  checkForCgError("loading fragment program");

	// Get Handles To Each Of Our Parameters So That
	// We Can Change Them At Will Within Our Code
	position		= cgGetNamedParameter(myCgVertexProgram, "position");
	texCoord			= cgGetNamedParameter(myCgVertexProgram, "texCoord");
	oPosition	= cgGetNamedParameter(myCgVertexProgram, "oPosition");
	oTexCoord = cgGetNamedParameter(myCgVertexProgram, "oTexCoord");
	modelViewProj	= cgGetNamedParameter(myCgVertexProgram, "modelViewProj");
	lightMapTexCoord = cgGetNamedParameter(myCgFragmentProgram, "lightMapTexCoord");
	color = cgGetNamedParameter(myCgFragmentProgram, "color");
	
	lightMap = cgGetNamedParameter(myCgFragmentProgram, "lightMap");
	checkForCgError("getting lightMap parameter");

	cgGLSetTextureParameter(lightMap,LIGHT_TEX_NAME);
	checkForCgError("setting light 2D texture");
  
	decalMap = cgGetNamedParameter(myCgFragmentProgram, "decalMap");
	checkForCgError("getting decalMap parameter");

	cgGLSetTextureParameter(decalMap,BASE_TEX_NAME);
	checkForCgError("setting decal 2D texture");

	return true;														// Return true (Initialization Successful)
}


//helper functions for working with vectors
Vector3 cross(const Vector3& a, const Vector3& b) {

	Vector3 c;
	c.x = a.y*b.z - a.z*b.y;
	c.y = a.z*b.x - a.x*b.z;
	c.z = a.x*b.y - a.y*b.x;

	return(c);
}

float dot(const Vector3& a, const Vector3& b) {
	return(a.x*b.x + a.y*b.y + a.z*b.z);
}

float length(const Vector3& a) {
	return(sqrt(dot(a, a)));
}

Vector3 normalize(const Vector3& a) {
	Vector3 u;
	float mag_a = length(a);
	u.x = a.x/mag_a;
	u.y = a.y/mag_a;
	u.z = a.z/mag_a;
	return(u);
}


/*Draw ground plane from ATI example - fairly straight forward - this is for shader */
void DrawGroundPlane(void)
{
	float size = 12.0f;
	float altitude = -0.001f;

	glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0.0f, 1.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-size, altitude, -size);

		glTexCoord2f(0.0f, 0.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-size, altitude, size);

		glTexCoord2f(1.0f, 1.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(size, altitude, -size);

		glTexCoord2f(1.0f, 0.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(size, altitude, size);
	glEnd();
}

/*Draw ground plane from ATI example - fairly straight forward - this is for multi-texturing */
void DrawMultitextureGroundPlane(float offset)
{
	float size = 12.0f;
	float altitude = -0.001f;

	glBegin(GL_TRIANGLE_STRIP);
		glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 1.0f);
		glMultiTexCoord2f(GL_TEXTURE1_ARB, 0.0f+offset, 1.0f+offset);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-size, altitude, -size);

		glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 0.0f);
		glMultiTexCoord2f(GL_TEXTURE1_ARB, 0.0f+offset, 0.0f+offset);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-size, altitude, size);

		glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 1.0f);
		glMultiTexCoord2f(GL_TEXTURE1_ARB, 1.0f+offset, 1.0f+offset);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(size, altitude, -size);

		glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 0.0f);
		glMultiTexCoord2f(GL_TEXTURE1_ARB, 1.0f+offset, 0.0f+offset);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(size, altitude, size);
	glEnd();
}


/*Display function */
void display() {

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
		//set up the camera
		gluLookAt(eye.x, eye.y, eye.z, la.x, la.y, la.z, 0, 1,0);
		glPushMatrix();
			
			if(gbMultitexturing)
			{
				cout << "using multitexturing" << endl;
				glActiveTexture(GL_TEXTURE0_ARB);
				glBindTexture(GL_TEXTURE_2D, BASE_TEX_NAME);
				glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
				glEnable(GL_TEXTURE_2D);

				glActiveTexture(GL_TEXTURE1_ARB);
				glBindTexture(GL_TEXTURE_2D, LIGHT_TEX_NAME);
				glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
				glEnable(GL_TEXTURE_2D);

				DrawMultitextureGroundPlane(gOffset);
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_TEXTURE_2D);
			}
			else
			{	
				cout << "using a shader" << endl;
				cgGLSetStateMatrixParameter(modelViewProj, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
			
				cgGLBindProgram(myCgVertexProgram);
				checkForCgError("binding vertex program");

				cgGLEnableProfile(myCgVertexProfile);
				checkForCgError("enabling vertex profile");

				cgGLBindProgram(myCgFragmentProgram);
				checkForCgError("binding fragment program");

				cgGLEnableProfile(myCgFragmentProfile);
				checkForCgError("enabling fragment profile");
			  
				cgGLEnableTextureParameter(lightMap);
				checkForCgError("enable light texture");
				
				cgGLEnableTextureParameter(decalMap);
				checkForCgError("enable decal texture");
			  
				DrawGroundPlane();
				
				cgGLDisableProfile(myCgVertexProfile);
				checkForCgError("disabling vertex profile");

				cgGLDisableProfile(myCgFragmentProfile);
				checkForCgError("disabling fragment profile");
			  
				cgGLDisableTextureParameter(lightMap);
				checkForCgError("disabling light texture");
			
				cgGLDisableTextureParameter(decalMap);
				checkForCgError("disabling decal texture");
			  
			}
		glPopMatrix();
	glPopMatrix();
	glutSwapBuffers();

}

void mouse(int btn, int state, int x, int y){
  startx = x;
  starty = GH-y-1;
}


/*Used to rotate view */
void mouseMove(int x, int y) {
  Vector3 diff, newV;

  endx = x;
  endy = GH-y-1;
  
  theta += ((endx - startx)/(float)GW)*90.0;
  phi += ((endy - starty)/(float)GH)*90.0; 
  
  startx = endx; starty = endy;
  newV.x = cosf(phi*(3.14/180.0))*cosf(theta*(3.14/180.0));
  newV.y = -1.0*sinf(phi*(3.14/180.0));
  newV.z = 1.0*cosf(phi*(3.14/180.0))*cosf((90.0-theta)*(3.14/180.0));
  diff.x = (la.x - eye.x) - newV.x;
  diff.y = (la.y - eye.y) - newV.y;
  diff.z = (la.z - eye.z) - newV.z;
  la.x = la.x - diff.x;
  la.y = la.y - diff.y;
  la.z = la.z - diff.z;

  glutPostRedisplay();
}

/* ASWD camera controls */
void keyboard(unsigned char key, int x, int y) {

  float delta = 0.2;
  Vector3 gaze,left;
  switch(key) {
	  case 'a':
		gaze.x = la.x - eye.x;
		gaze.y = la.y - eye.y;
		gaze.z = la.z - eye.z;
		left = cross(Vector3(0, 1, 0), gaze);
		eye.x += delta*left.x;
		eye.y += delta*left.y;
		eye.z += delta*left.z;
		la.x += delta*left.x;
		la.y += delta*left.y;
		la.z += delta*left.z;
		break;
	  case 'd':
		gaze.x = la.x - eye.x;
		gaze.y = la.y - eye.y;
		gaze.z = la.z - eye.z;
		left = cross(Vector3(0, 1, 0), gaze);
		eye.x -= delta*left.x;
		eye.y -= delta*left.y;
		eye.z -= delta*left.z;
		la.x -= delta*left.x;
		la.y -= delta*left.y;
		la.z -= delta*left.z;
		glutPostRedisplay();
		break;
	  case 's':
		gaze.x = la.x - eye.x;
		gaze.y = la.y - eye.y;
		gaze.z = la.z - eye.z;
		eye.x -= delta*gaze.x;
		eye.y -= delta*gaze.y;
		eye.z -= delta*gaze.z;
		la.x -= delta*gaze.x;
		la.y -= delta*gaze.y;
		la.z -= delta*gaze.z;
		glutPostRedisplay();
		break;
	   case 'w':
		gaze.x = la.x - eye.x;
		gaze.y = la.y - eye.y;
		gaze.z = la.z - eye.z;
		eye.x += delta*gaze.x;
		eye.y += delta*gaze.y;
		eye.z += delta*gaze.z;
		la.x += delta*gaze.x;
		la.y += delta*gaze.y;
		la.z += delta*gaze.z;
		glutPostRedisplay();
		break;
	  case 'q':
		  exit( EXIT_SUCCESS );
		  break;
	  case 'l':
		  gbMultitexturing = !gbMultitexturing;
		  break;
	  default:
		  break;
	}
	glutPostRedisplay();
}

void reshape(int w, int h) {
	GW = w;
	GH = h;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(60.0, (float)w/h, 0.1, 50);
	glMatrixMode(GL_MODELVIEW);
	glViewport(0, 0, w, h);

	glutPostRedisplay();
}


int main( int argc, char** argv ) {

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(500, 500); 
	glutInitWindowPosition(100, 100);
	glutCreateWindow("Light map example");
	glClearColor(1.0, 1.0, 1.0, 1.0);
	
	/*register glut call backs */
	glutDisplayFunc( display );
	glutReshapeFunc( reshape );
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(mouseMove);
	
	/* projection/modelview set up */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, 1.0, 0.1, 50);
	glMatrixMode(GL_MODELVIEW);
	glViewport(0, 0, 500, 500); //image to pixel

    /*initialize camera stuff */
	GW=GH = 500;
	phi = 0.0; theta = -90.0;	
	eye.x = 0.0; eye.y = 2.0; eye.z = 2.0;
	la.x = 0; la.y = 2.0; la.z = 0;
	/* Start with the shader - this variable is defined in tex.h */
	gbMultitexturing = 0;
  
	/* Load in the two textures we'll use for this program and create texture objects */
	LoadTexture("texture/image1.bmp",    BASE_TEX_NAME);	// Here we load a normal texture
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	LoadTexture("texture/lightmap.bmp", LIGHT_TEX_NAME);	// Here we load a lightmap
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	
	/*set up CG after we've loaded in the textures*/
	Initialize ();

	glutMainLoop();
}


